package pku.edu.hocra;

import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import pku.edu.hocra.graph.ExplodedGraphBuilder;
import pku.edu.hocra.graph.ds.Graph;
import pku.edu.hocra.resource.Constants;
import pku.edu.hocra.resource.FileManager;
import soot.MethodOrMethodContext;
import soot.PackManager;
import soot.Scene;
import soot.SootClass;
import soot.SootMethod;
import soot.jimple.toolkits.callgraph.CHATransformer;
import soot.jimple.toolkits.callgraph.Edge;
import soot.jimple.toolkits.callgraph.ReachableMethods;
import soot.jimple.toolkits.pointer.DumbPointerAnalysis;
import soot.util.Chain;
import soot.util.HashChain;
import soot.util.queue.QueueReader;

public class DriverLibrary {
	
	public static void main(String[] args) {
		
		FileManager.makeDir("dddata");
		FileManager.makeDir("dddata/library");
		
		initializeSootClassPath();

		List<SootClass> libraryClasses = getLibraryClasses();
		List<SootMethod> libraryEntryMethods = getMethodsFromClasses(libraryClasses);
		System.out.println("|LibCls|=" + libraryClasses.size());
		System.out.println("|LibEntryMtd|=" + libraryEntryMethods.size());
		List<MethodOrMethodContext> entryPoints = new ArrayList<MethodOrMethodContext>();
		entryPoints.addAll(libraryEntryMethods);
		
		CHATransformer.v().transform();
		
		ReachableMethods reachableMethods = Scene.v().getReachableMethods();
		Set<SootMethod> libraryMethods = new HashSet<SootMethod>();
		QueueReader<MethodOrMethodContext> qr = reachableMethods.listener();
		while (qr.hasNext()) {
			SootMethod sm = qr.next().method();
			if (!sm.getDeclaringClass().toString().startsWith("sun."))
				libraryMethods.add(sm);
		}
		System.out.println("|LibMtd|=" + libraryMethods.size());
		System.out.println("|LibInterpEdge|=" + Scene.v().getCallGraph().size());
		try {
			ExplodedGraphBuilder egb = new ExplodedGraphBuilder(Scene.v().getCallGraph(), libraryMethods);
			egb.build("dddata/library/LibraryMethod.dat", 
					"dddata/library/LibraryGraph.dat", 
					"dddata/library/LibraryCG.dat",
					"dddata/library/LibraryMaxNumber.dat");
		} catch (FileNotFoundException e1) {
			e1.printStackTrace();
		}
		
	}
	
	private static void initializeSootClassPath() {
		String jreLibPrefix = Constants.jrePath; 
		String sp = Constants.sep;
		String allClassPath = 
					jreLibPrefix+"rt.jar"					
					+sp+jreLibPrefix+"jce.jar";
		Scene.v().setSootClassPath(allClassPath);
	}
	
	private static List<SootClass> getLibraryClasses() {
		Scene.v().loadBasicClasses();
		Scene.v().loadDynamicClasses();
		List<SootClass> allLibraryClasses = new ArrayList<SootClass>();
		HashSet<SootClass> processedClasses = new HashSet<SootClass>();
		for (boolean isDone = false; !isDone; ) {
			isDone = true;
			Chain<SootClass> csc = new HashChain<SootClass>();
			csc.addAll(Scene.v().getLibraryClasses());
			for (SootClass sc : csc) {
				if (!processedClasses.contains(sc)) {
					processedClasses.add(sc);
					isDone = false;
					SootClass newClass = Scene.v().loadClassAndSupport(sc.getName());
					allLibraryClasses.add(newClass);
				}
			}
		}
		return allLibraryClasses;
	}
	
	private static List<SootMethod> getMethodsFromClasses(List<SootClass> allClasses) {
		List<SootMethod> result = new ArrayList<SootMethod>();
		for (SootClass libraryClass: allClasses) {
			result.addAll(libraryClass.getMethods());
		}
		return result;
	}

}
