package pku.edu.hocra;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Scanner;
import java.util.Set;

import pku.edu.hocra.graph.ds.IntPair;
import pku.edu.hocra.graph.ds.MultiMap;
import pku.edu.hocra.graph.ds.MutualMap;
import pku.edu.hocra.resource.Constants;

class TriInt {
	int m1, u1, m2;
	public TriInt(int m1, int u1, int m2) {
		this.m1 = m1;
		this.u1 = u1;
		this.m2 = m2;
	}
	@Override
	public boolean equals(Object o) {
		if (o instanceof TriInt) {
			TriInt tmp = (TriInt) o;
			return m1 == tmp.m1 && u1 == tmp.u1 && m2 == tmp.m2;			
		}
		return false;
	}
	@Override
	public int hashCode() {
		return ( (m1 * 109 + u1) * 1009 + m2 ) * 49999;
	}
}

public class DriverReindex {
	
	MultiMap<IntPair, String> libE2L = new MultiMap<IntPair, String>();
	Map<Integer, String> libI2M = new HashMap<Integer, String>();
	Map<String, Integer> libM2B = new HashMap<String, Integer>();
	Map<String, Integer> libM2Size = new HashMap<String, Integer>();
	MutualMap<String, Integer> libM2Mi = new MutualMap<String, Integer>();
	Map<Integer, String> sbjI2M = new HashMap<Integer, String>();
	Map<String, Integer> sbjM2B = new HashMap<String, Integer>();
	MutualMap<String, Integer> sbjM2Mi = new MutualMap<String, Integer>();
	Map<TriInt, Integer> libMUM2Pi = new HashMap<TriInt, Integer>();
	int n = 0;
	int nLib = 0;
	
	public static void main(String[] args) throws IOException {
		for (String subjectName : Constants.subjectNames) {
			System.out.println("Running <" + subjectName + ">");
			new DriverReindex().run(subjectName);
		}
		new DriverReindex().modifyLibrary();
	}
	
	public void modifyLibrary() throws IOException {
		
		FileInputStream fis = new FileInputStream("dddata/library/LibraryMaxNumber.dat");
		Scanner sc = new Scanner(fis);
		nLib = sc.nextInt();
		sc.close();
		
		fis = new FileInputStream("dddata/library/LibraryGraph.dat");
		FileOutputStream fos = new FileOutputStream("dddata/library/LibraryGraph.tmp");
		FileOutputStream fos2 = new FileOutputStream("dddata/library/LibraryGraphCHA.dat");
		
		PrintStream ps = new PrintStream(fos);
		PrintStream ps2 = new PrintStream(fos2);
		sc = new Scanner(fis);
		
		ps.println(nLib);
		ps2.println(nLib);
		while (sc.hasNextLine()) {
			String line = sc.nextLine();
			ps.println(line);
			ps2.println(line.replaceAll("V", ""));
		}
		
		sc.close();
		ps.close();
		ps2.close();
		
		new File("dddata/library/LibraryGraph.dat").delete();
		new File("dddata/library/LibraryGraph.tmp").renameTo(new File("dddata/library/LibraryGraph.dat"));
		
		fis = new FileInputStream("dddata/library/LibraryMethod.dat");
		fos = new FileOutputStream("dddata/library/LibraryMethodInfo.dat");
		
		sc = new Scanner(fis);
		ps = new PrintStream(fos);
		
		while (sc.hasNextLine()) {
			String line = sc.nextLine();
			ps.println(
					line.split(";")[1]
				+ 	" "
				+	line.split(";")[2]
			);
		}
		
		sc.close();
		ps.close();
		
		
	}
	
	public void run(String subjectName) throws IOException {
		
		FileInputStream fis = new FileInputStream("dddata/library/LibraryMaxNumber.dat");
		Scanner sc = new Scanner(fis);
		nLib = sc.nextInt();
		sc.close();
		
		fis = new FileInputStream("dddata/library/LibraryGraph.dat");
		sc = new Scanner(fis);
		int maxp = 0;
		while (sc.hasNextLine()) {
			String line = sc.nextLine();
			String[] ls = line.split(" ");
			IntPair e = new IntPair(Integer.valueOf(ls[0]), Integer.valueOf(ls[1]));
			if (!(ls.length == 2 || ls[2].equals("C"))) {
				libE2L.put(e, ls[2] + " " + ls[3]);
			}
			if (ls.length > 2 && (ls[2].equals("L") || ls[2].equals("R")))
				if (Integer.valueOf(ls[3]) > maxp)
					maxp = Integer.valueOf(ls[3]);
		}
		sc.close();
		
		n = nLib + 2;

		fis = new FileInputStream("dddata/library/LibraryMethod.dat");
		sc = new Scanner(fis);
		while (sc.hasNextLine()) {
			String line = sc.nextLine();
			String[] ls = line.split(";");
			if (ls.length != 4)
				System.err.println(line);
			int start = Integer.valueOf(ls[1]);
			int len = Integer.valueOf(ls[2]);
			libM2B.put(ls[0], start);
			libM2Size.put(ls[0], len);
			for (int i = start; i < start+len; i++)
				libI2M.put(i, ls[0]);
			libM2Mi.put(ls[0], Integer.valueOf(ls[3]));
		}
		sc.close();
		
		fis = new FileInputStream("dddata/subjects/" + subjectName + "/SubjectMethod.dat");
		sc = new Scanner(fis);
		while (sc.hasNextLine()) {
			String line = sc.nextLine();
			String[] ls = line.split(";");
			if (ls.length != 4)
				System.out.println(line);
			int start = Integer.valueOf(ls[1]);
			int len = Integer.valueOf(ls[2]);
			sbjM2B.put(ls[0], start);
			//if (libM2Size.containsKey(ls[0]))
			//	if (libM2Size.get(ls[0]) != len)
			//		System.out.println("<>" + ls[0]);
			for (int i = start; i < start+len; i++)
				sbjI2M.put(i, ls[0]);
			sbjM2Mi.put(ls[0], Integer.valueOf(ls[3]));
		}
		sc.close();
		
		FileOutputStream fos = new FileOutputStream("dddata/subjects/" + subjectName + "/TrueSubjectGraph.dat");
		PrintStream ps = new PrintStream(fos);
		
		fis = new FileInputStream("dddata/subjects/" + subjectName + "/SubjectGraph.dat");
		sc = new Scanner(fis);
		while (sc.hasNextLine()) {
			String line = sc.nextLine();
			String[] ls = line.split(" ");
			IntPair e = new IntPair(Integer.valueOf(ls[0]), Integer.valueOf(ls[1]));
			int ui = process(e.u);
			int vi = process(e.v);
			//IntPair ee = new IntPair(ui, vi);
			if (ui > nLib || vi > nLib) {
				if (ls.length == 2)
					ps.println(ui + " " + vi);
				else if (ls[2].charAt(0) == 'L' || ls[2].charAt(0) == 'R') {
					ps.println(ui + " " + vi + " " + ls[2].charAt(0) + " " + (Integer.valueOf(ls[3])+maxp));
				}
			}
		}
		sc.close();
		
		fis = new FileInputStream("dddata/library/LibraryGraph.dat");
		sc = new Scanner(fis);
		while (sc.hasNextLine()) {
			String line = sc.nextLine();
			String[] ls = line.split(" ");
			if (ls.length > 2) continue;
			IntPair e = new IntPair(Integer.valueOf(ls[0]), Integer.valueOf(ls[1]));
			String uMethod = libI2M.get(e.u);
			String vMethod = libI2M.get(e.v);
			if (uMethod.equals(vMethod) && sbjM2B.containsKey(uMethod)) {
				ps.println(e.u + " " + e.v);
			}
		}
		sc.close();
		
		fis = new FileInputStream("dddata/library/LibraryCG.dat");
		sc = new Scanner(fis);
		while (sc.hasNextLine()) {
			String line = sc.nextLine();
			String[] ls = line.split(" ");
			TriInt ti = new TriInt(Integer.valueOf(ls[0]), Integer.valueOf(ls[1]), Integer.valueOf(ls[2]));
			int pi = Integer.valueOf(ls[4]);
			libMUM2Pi.put(ti, pi);
		}
		sc.close();
		
		Set<Integer> piSet = new HashSet<Integer>();
		fis = new FileInputStream("dddata/subjects/" + subjectName + "/SubjectCG.dat");
		sc = new Scanner(fis);
		PrintStream psVirtualCall = new PrintStream(new FileOutputStream("dddata/subjects/" + subjectName + "/ImplementedVirtualCall.dat"));
		while (sc.hasNextLine()) {
			String line = sc.nextLine();
			String[] ls = line.split(" ");
			int m1sub = Integer.valueOf(ls[0]);
			int m2sub = Integer.valueOf(ls[2]);
			String m1name = sbjM2Mi.getKey(m1sub);
			String m2name = sbjM2Mi.getKey(m2sub);
			if (libM2Mi.containsKey(m1name) && libM2Mi.containsKey(m2name)) {
				int m1lib = libM2Mi.getValue(m1name);
				int m2lib = libM2Mi.getValue(m2name);
				TriInt ti = new TriInt(m1lib, Integer.valueOf(ls[1]), m2lib);
				if (!libMUM2Pi.containsKey(ti)) {
					System.err.println(m1name + " " + m2name);
				}
				else {
					piSet.add(libMUM2Pi.get(ti));
					if (ls[3].equals("V"))
						psVirtualCall.println(libMUM2Pi.get(ti));
				}
			}
		}
		sc.close();
		
		fis = new FileInputStream("dddata/library/LibraryGraph.dat");
		sc = new Scanner(fis);
		
		while (sc.hasNextLine()) {
			String line = sc.nextLine();
			String[] ls = line.split(" ");
			if (ls.length != 4) continue;
			if (ls[2].equals("C")) continue;
			int pi = Integer.valueOf(ls[3]);
			if (piSet.contains(pi)) {
				IntPair e = new IntPair(Integer.valueOf(ls[0]), Integer.valueOf(ls[1]));
				ps.println(e.u + " " + e.v + " " + ls[2].charAt(0) + " " + pi);
			}
		}
		sc.close();

		ps.close();
		psVirtualCall.close();
		
		fis = new FileInputStream("dddata/subjects/" + subjectName + "/TrueSubjectGraph.dat");
		sc = new Scanner(fis);
		int maxNode = 0;
		while (sc.hasNextLine()) {
			String line = sc.nextLine();
			String[] ls = line.split(" ");
			for (int i = 0; i < 2; i++) {
				if (Integer.valueOf(ls[i]) > maxNode) {
					maxNode = Integer.valueOf(ls[i]);
				}
			}
		}
		sc.close();
		maxNode ++;
		
		fis = new FileInputStream("dddata/subjects/" + subjectName + "/TrueSubjectGraph.dat");
		sc = new Scanner(fis);
		fos = new FileOutputStream("dddata/subjects/" + subjectName + "/TrueSubjectGraph.tmp");
		ps = new PrintStream(fos);
		ps.println(maxNode);
		while (sc.hasNextLine()) {
			String line = sc.nextLine();
			ps.println(line);
		}
		sc.close();
		ps.close();
		
		new File("dddata/subjects/" + subjectName + "/TrueSubjectGraph.dat").delete();
		new File("dddata/subjects/" + subjectName + "/TrueSubjectGraph.tmp")
			.renameTo(new File("dddata/subjects/" + subjectName + "/TrueSubjectGraph.dat"));
		
	}
	
	private Map<Integer, Integer> indexMapping = new HashMap<Integer, Integer>();

	private int process(int i) {
		if (indexMapping.containsKey(i)) {
			return indexMapping.get(i);
		}
		String sbj_m = sbjI2M.get(i);
		if (libM2B.containsKey(sbj_m)) {
			// It is a library method
			int sbj_b = sbjM2B.get(sbj_m);
			int lib_b = libM2B.get(sbj_m);
			int new_i = i-sbj_b+lib_b;
			indexMapping.put(i, new_i);
			return new_i;
		}
		else {
			++n;
			indexMapping.put(i, n);
			return n;
		}
	}
	
}
