//
// Created by Wayne on 16/1/23.
//

#include "analyze_tal.h"

using std::map;
using std::make_pair;

unordered_set<int> special;

static inline void add(int sym, queue<const edge*> &wl, application_graph *gr, int src, int tgt) {
    if (!gr->is_redundant(src, tgt, sym)) {
        auto new_edge = new edge(src, tgt, sym);
        wl.push(new_edge);
        gr->add_record(new_edge);
    }
}

static inline void addS(queue<const edge*> &wl, application_graph *gr, int src, int tgt) {
    add(GRAM_S, wl, gr, src, tgt);
}

static inline void addL(queue<const edge*> &wl, application_graph *gr, int src, int tgt) {
    add(GRAM_L, wl, gr, src, tgt);
}

static inline void addR(queue<const edge*> &wl, application_graph *gr, int src, int tgt) {
    add(GRAM_R, wl, gr, src, tgt);
}

static inline void addD(queue<const edge*> &wl, application_graph *gr, int src, int tgt) {
    add(GRAM_D, wl, gr, src, tgt);
}

static inline void add_fst(int sym, queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt) {
    if (!gr->is_fst_redundant(src, tgt, sym)) {
        auto new_edge = new tal_fst_edge(sym, src, tgt);
        wl.push(new_edge);
        gr->add_fst_record(new_edge);
    }
}

static inline void add_fst_S(queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt) {
    add_fst(GRAM_S, wl, gr, src, tgt);
}

static inline void add_fst_L(queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt) {
    add_fst(GRAM_L, wl, gr, src, tgt);
}

static inline void add_fst_R(queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt) {
    add_fst(GRAM_R, wl, gr, src, tgt);
}

static inline void add_fst_D(queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt) {
    add_fst(GRAM_D, wl, gr, src, tgt);
}

static inline void add_snd(int sym, queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt, int msrc, int mtgt) {
    if (!gr->is_fst_redundant(src, tgt, sym) && special.count(src) != 0 && special.count(tgt) != 0 && special.count(msrc) != 0 && special.count(mtgt) != 0 && !gr->is_snd_redundant(src, tgt, msrc, mtgt, sym)) {
        auto new_edge = new tal_snd_edge(sym, src, tgt, msrc, mtgt);
        wl.push(new_edge);
        gr->add_snd_record(new_edge);
    }
}

static inline void add_snd_S(queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt, int msrc, int mtgt) {
    add_snd(GRAM_S, wl, gr, src, tgt, msrc, mtgt);
}

static inline void add_snd_L(queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt, int msrc, int mtgt) {
    add_snd(GRAM_L, wl, gr, src, tgt, msrc, mtgt);
}

static inline void add_snd_R(queue<const tal_edge_base*> &wl, tal_library_graph *gr, int src, int tgt, int msrc, int mtgt) {
    add_snd(GRAM_R, wl, gr, src, tgt, msrc, mtgt);
}

const application_graph* analyze_tal(const application_graph *app_gr, tal_library_graph *lib_gr, Logger &logger) {
    auto start = clock();

    unordered_set<int> inputs, outputs;
    unordered_set<int> actual_nodes;
    int lib_nodes = lib_gr->get_num_nodes();
    for (auto e : app_gr->all_parens()) {
        int src = e->get_source();
        int tgt = e->get_target();
        if (src < lib_nodes && tgt >= lib_nodes) {
            outputs.insert(src);
        }
        if (tgt < lib_nodes && src >= lib_nodes) {
            inputs.insert(tgt);
        }
        actual_nodes.insert(src);
        actual_nodes.insert(tgt);
    }
    for (auto e : app_gr->all_edges()) {
        actual_nodes.insert(e->get_source());
        actual_nodes.insert(e->get_target());
    }

    unordered_set<int> keys;
    for (int n : inputs) keys.insert(n);
    for (int n : outputs) keys.insert(n);
    for (int n : lib_gr->get_chain_nodes()) {
        if (actual_nodes.count(n) != 0) {
            keys.insert(n);
        }
    }

    cerr << "Keys: " << keys.size() << endl;

    queue<const edge*> work_list;
    application_graph *tmp_gr = new application_graph(app_gr->get_num_nodes());
    tal_library_graph *the_lib = new tal_library_graph(app_gr->get_num_nodes());

    for (auto e : app_gr->all_parens()) {
        if (e->get_source() >= lib_nodes || e->get_target() >= lib_nodes) {
            tmp_gr->add_paren(e);
        }
    }
    for (int src = 0; src < app_gr->get_num_nodes(); src++) {
        for (auto e : app_gr->edges_from(src, GRAM_S0)) {
            if (src >= lib_nodes || e->get_target() >= lib_nodes) {
                addS(work_list, tmp_gr, e->get_source(), e->get_target());
            }
        }
    }
    for (int src : keys) {
        for (int s = 0; s < 7; s++) {
            if (s == GRAM_S0 || s == GRAM_L0 || s == GRAM_R0) continue;
            for (auto fe : lib_gr->fst_edges_from(src, s)) {
                if (keys.count(fe->get_target()) != 0) {
                    add(fe->get_symbol(), work_list, tmp_gr, fe->get_source(), fe->get_target());
                }
            }
        }
    }
    /*for (auto e : lib_gr->all_parens()) {
        if (keys.count(e->get_source()) != 0 && keys.count(e->get_target()) != 0) {
            tmp_gr->add_paren(new tal_fst_edge(e->get_symbol(), e->get_source(), e->get_target()));
        }
    }*/
    special.clear();
    for (auto fe : tmp_gr->all_parens()) {
        if (is_left_paren(fe->get_symbol())) {
            addL(work_list, tmp_gr, fe->get_source(), fe->get_target());
        } else if (is_right_paren(fe->get_symbol())) {
            addR(work_list, tmp_gr, fe->get_source(), fe->get_target());
        } else {
            throw "bad implementation";
        }
        special.insert(fe->get_source());
        special.insert(fe->get_target());
    }
    vector<map<int, vector<pair<int, int>>>> record;
    record.resize(tmp_gr->get_num_nodes());
    auto &lefts = tmp_gr->get_left_parens();
    auto &rights = tmp_gr->get_right_parens();
    for (auto it = lefts.begin(); it != lefts.end(); it++) {
        int left_paren = it->first;
        int right_paren = -it->first;
        if (rights.count(right_paren) != 0) {
            for (auto e1 : lefts.at(left_paren)) {
                for (auto e2 : rights.at(right_paren)) {
                    record[e1->get_target()][e2->get_source()].push_back(make_pair(e1->get_source(), e2->get_target()));
                }
            }
        }
    }
    for (int n = 0; n < tmp_gr->get_num_nodes(); n++) {
        if (record[n].count(n) != 0) {
            auto &pairs = record[n][n];
            for (auto &p : pairs) {
                addS(work_list, tmp_gr, p.first, p.second);
            }
        }
    }
    for (int src : keys) {
        for (int s = 0; s < 3; s++) {
            for (auto se : lib_gr->snd_edges_from(src, s)) {
                if (keys.count(se->get_target()) != 0) {
                    if (special.count(se->get_source()) != 0 && special.count(se->get_target()) != 0 && special.count(se->get_mid_source()) != 0 && special.count(se->get_mid_target()) != 0) {
                        the_lib->add_snd_edge(se);
                    }
                }
            }
        }
    }

    cerr << "SND edges: " << the_lib->get_num_edges() << endl;

    int steps = 0;
    while (!work_list.empty()) {
        if (steps % 100000 == 0) {
            cerr << "Steps: " << steps << ", Worklist: " << work_list.size() << endl;
        }
        steps += 1;
        auto e = work_list.front();
        work_list.pop();
        tmp_gr->add_edge(e);

        int source = e->get_source();
        int target = e->get_target();
        int symbol = e->get_symbol();

        if (symbol == GRAM_S) {
            for (auto se : the_lib->snd_edges_of_mid(source, target, GRAM_S)) {
                addS(work_list, tmp_gr, se->get_source(), se->get_target());
            }
            for (auto se : the_lib->snd_edges_of_mid(source, target, GRAM_L)) {
                addL(work_list, tmp_gr, se->get_source(), se->get_target());
            }
            for (auto se : the_lib->snd_edges_of_mid(source, target, GRAM_R)) {
                addR(work_list, tmp_gr, se->get_source(), se->get_target());
            }
            // S -> @SS
            for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                addS(work_list, tmp_gr, source, forward->get_target());
            }
            // S -> S@S
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                addS(work_list, tmp_gr, backward->get_source(), target);
            }
            // L -> L@S
            for (auto backward : tmp_gr->edges_to(source, GRAM_L)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }
            // L -> @SL
            for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }
            // R -> R@S
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }
            // R -> @SR
            for (auto forward : tmp_gr->edges_from(target, GRAM_R)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }
            // D -> @SD
            for (auto forward : tmp_gr->edges_from(target, GRAM_D)) {
                addD(work_list, tmp_gr, source, forward->get_target());
            }
            // D -> D@S
            for (auto backward : tmp_gr->edges_to(source, GRAM_D)) {
                addD(work_list, tmp_gr, backward->get_source(), target);
            }
            // S -> {i@S}i
            if (record[source].count(target) != 0) {
                auto &pairs = record[source][target];
                for (auto &p : pairs) {
                    addS(work_list, tmp_gr, p.first, p.second);
                }
            }
        } else if (symbol == GRAM_L) {
            // L -> @LL
            for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }
            // L -> L@L
            for (auto backward : tmp_gr->edges_to(source, GRAM_L)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }
            // L -> @LS
            for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }
            // L -> S@L
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }
            // D -> R@L
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                addD(work_list, tmp_gr, backward->get_source(), target);
            }
            // D -> D@L
            for (auto backward : tmp_gr->edges_to(source, GRAM_D)) {
                addD(work_list, tmp_gr, backward->get_source(), target);
            }
        } else if (symbol == GRAM_R) {
            // R -> @RR
            for (auto forward : tmp_gr->edges_from(target, GRAM_R)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }
            // R -> R@R
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }
            // R -> @RS
            for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }
            // R -> S@R
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }
            // D -> @RL
            for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                addD(work_list, tmp_gr, source, forward->get_target());
            }
            // D -> @RD
            for (auto forward : tmp_gr->edges_from(target, GRAM_D)) {
                addD(work_list, tmp_gr, source, forward->get_target());
            }
        } else if (symbol == GRAM_D) {
            // D -> @DS
            for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                addD(work_list, tmp_gr, source, forward->get_target());
            }
            // D -> S@D
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                addD(work_list, tmp_gr, backward->get_source(), target);
            }
            // D -> @DL
            for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                addD(work_list, tmp_gr, source, forward->get_target());
            }
            // D -> R@D
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                addD(work_list, tmp_gr, backward->get_source(), target);
            }
        } else {
            throw "bad implementation";
        }
    }

    auto end = clock();
    cerr << "Analyze time: " << (end - start) * 1.0 / CLOCKS_PER_SEC << endl;

    logger.log("Time of Analysis: ");
    logger.log((end - start) * 1.0 / CLOCKS_PER_SEC);
    logger.log("\n");

    return tmp_gr;
}