//
// Created by Wayne on 16/1/21.
//

#include "analyze_whole.h"

using std::queue;
using std::map;
using std::make_pair;

static inline void add(int symbol, queue<const edge*> &wl, application_graph *gr, int src, int tgt) {
    //if (src == tgt) return;
    if (!gr->is_redundant(src, tgt, symbol)) {
        auto new_edge = new edge(src, tgt, symbol);
        wl.push(new_edge);
        gr->add_record(new_edge);
    }
}

static inline void addS(queue<const edge*> &wl, application_graph *gr, int src, int tgt) {
    add(GRAM_S, wl, gr, src, tgt);
}

static inline void addL(queue<const edge*> &wl, application_graph *gr, int src, int tgt) {
    add(GRAM_L, wl, gr, src, tgt);
}

static inline void addR(queue<const edge*> &wl, application_graph *gr, int src, int tgt) {
    add(GRAM_R, wl, gr, src, tgt);
}

static inline void addS0(queue<const edge*> &wl, application_graph *gr, int src, int tgt) {
    add(GRAM_S0, wl, gr, src, tgt);
}

static inline void addL0(queue<const edge*> &wl, application_graph *gr, int src, int tgt) {
    add(GRAM_L0, wl, gr, src, tgt);
}

static inline void addR0(queue<const edge*> &wl, application_graph *gr, int src, int tgt) {
    add(GRAM_R0, wl, gr, src, tgt);
}

static inline void addD(queue<const edge*> &wl, application_graph *gr, int src, int tgt) {
    add(GRAM_D, wl, gr, src, tgt);
}

const application_graph* analyze_whole(const application_graph *app_gr, Logger &logger) {
    auto start = clock();

    queue<const edge*> work_list;
    application_graph *tmp_gr = new application_graph(app_gr->get_num_nodes());

    vector<map<int, vector<pair<int, int>>>> record;
    record.resize(tmp_gr->get_num_nodes());
    auto &lefts = app_gr->get_left_parens();
    auto &rights = app_gr->get_right_parens();
    for (auto it = lefts.begin(); it != lefts.end(); it++) {
        int left_paren = it->first;
        int right_paren = -it->first;
        if (rights.count(right_paren) != 0) {
            for (auto e1 : lefts.at(left_paren)) {
                for (auto e2 : rights.at(right_paren)) {
                    record[e1->get_target()][e2->get_source()].push_back(make_pair(e1->get_source(), e2->get_target()));
                }
            }
        }
    }

    for (int src = 0; src < tmp_gr->get_num_nodes(); src++) {
        for (auto e : app_gr->edges_from(src, GRAM_S0)) {
            addS(work_list, tmp_gr, e->get_source(), e->get_target());
        }
    }
    for (auto e : app_gr->all_parens()) {
        if (is_left_paren(e->get_symbol())) {
            addL(work_list, tmp_gr, e->get_source(), e->get_target());
        } else if (is_right_paren(e->get_symbol())) {
            addR(work_list, tmp_gr, e->get_source(), e->get_target());
        } else {
            throw "bad implementation";
        }
    }
    for (int n = 0; n < tmp_gr->get_num_nodes(); n++) {
        if (record[n].count(n) != 0) {
            auto &pairs = record[n][n];
            for (auto &p : pairs) {
                addS(work_list, tmp_gr, p.first, p.second);
            }
        }
    }

    int steps = 0;
    while (!work_list.empty()) {
        if (steps % 100000 == 0) {
            cerr << "Steps: " << steps << ", Worklist: " << work_list.size() << endl;
        }
        steps += 1;
        auto e = work_list.front();
        work_list.pop();
        tmp_gr->add_edge(e);

        int source = e->get_source();
        int target = e->get_target();
        int symbol = e->get_symbol();

        if (symbol == GRAM_S) {
            // S -> @SS
            for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                addS(work_list, tmp_gr, source, forward->get_target());
            }
            // S -> S@S
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                addS(work_list, tmp_gr, backward->get_source(), target);
            }
            // L -> @SL
            for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }
            // L -> L@S
            for (auto backward : tmp_gr->edges_to(source, GRAM_L)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }
            // R -> @SR
            for (auto forward : tmp_gr->edges_from(target, GRAM_R)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }
            // R -> R@S
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }
            // S -> {i@S}i
            if (record[source].count(target) != 0) {
                auto &pairs = record[source][target];
                for (auto &p : pairs) {
                    addS(work_list, tmp_gr, p.first, p.second);
                }
            }
            // D -> D@S
            for (auto backward : tmp_gr->edges_to(source, GRAM_D)) {
                addD(work_list, tmp_gr, backward->get_source(), target);
            }
            // D -> @SD
            for (auto forward : tmp_gr->edges_from(target, GRAM_D)) {
                addD(work_list, tmp_gr, source, forward->get_target());
            }
        } else if (symbol == GRAM_L) {
            // L -> @LL
            for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }
            // L -> L@L
            for (auto backward : tmp_gr->edges_to(source, GRAM_L)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }
            // L -> S@L
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }
            // L -> @LS
            for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }
            // D -> R@L
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                addD(work_list, tmp_gr, backward->get_source(), target);
            }
            // D -> D@L
            for (auto backward : tmp_gr->edges_to(source, GRAM_D)) {
                addD(work_list, tmp_gr, backward->get_source(), target);
            }
        } else if (symbol == GRAM_R) {
            // R -> @RR
            for (auto forward : tmp_gr->edges_from(target, GRAM_R)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }
            // R -> R@R
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }
            // R -> S@R
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }
            // R -> @RS
            for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }
            // D -> @RL
            for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                addD(work_list, tmp_gr, source, forward->get_target());
            }
            // D -> @RD
            for (auto forward : tmp_gr->edges_from(target, GRAM_D)) {
                addD(work_list, tmp_gr, source, forward->get_target());
            }
        } else if (symbol == GRAM_D) {
            // D -> @DL
            for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                addD(work_list, tmp_gr, source, forward->get_target());
            }
            // D -> R@D
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                addD(work_list, tmp_gr, backward->get_source(), target);
            }
            // D -> @DS
            for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                addD(work_list, tmp_gr, source, forward->get_target());
            }
            // D -> S@D
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                addD(work_list, tmp_gr, backward->get_source(), target);
            }
        } else {
            throw "bad implementation";
        }
    }

    auto end = clock();
    cerr << "Analyze time: " << (end - start) * 1.0 / CLOCKS_PER_SEC << endl;

    logger.log("Time of Analysis: ");
    logger.log((end - start) * 1.0 / CLOCKS_PER_SEC);
    logger.log("\n");

    return tmp_gr;
}

const application_graph* analyze_whole_opt(const application_graph *app_gr) {
    auto start = clock();

    queue<const edge*> work_list;
    application_graph *tmp_gr = new application_graph(app_gr->get_num_nodes());

    vector<map<int, vector<pair<int, int>>>> record;
    record.resize(tmp_gr->get_num_nodes());
    auto &lefts = app_gr->get_left_parens();
    auto &rights = app_gr->get_right_parens();
    for (auto it = lefts.begin(); it != lefts.end(); it++) {
        int left_paren = it->first;
        int right_paren = -it->first;
        if (rights.count(right_paren) != 0) {
            for (auto e1 : lefts.at(left_paren)) {
                for (auto e2 : rights.at(right_paren)) {
                    record[e1->get_target()][e2->get_source()].push_back(make_pair(e1->get_source(), e2->get_target()));
                }
            }
        }
    }

    for (int src = 0; src < tmp_gr->get_num_nodes(); src++) {
        for (auto e : app_gr->edges_from(src, GRAM_S0)) {
            addS0(work_list, tmp_gr, e->get_source(), e->get_target());
        }
    }
    for (auto e : app_gr->all_parens()) {
        if (is_left_paren(e->get_symbol())) {
            addL0(work_list, tmp_gr, e->get_source(), e->get_target());
        } else if (is_right_paren(e->get_symbol())) {
            addR0(work_list, tmp_gr, e->get_source(), e->get_target());
        } else {
            throw "bad implementation";
        }
    }
    for (int n = 0; n < tmp_gr->get_num_nodes(); n++) {
        if (record[n].count(n) != 0) {
            auto &pairs = record[n][n];
            for (auto &p : pairs) {
                addS0(work_list, tmp_gr, p.first, p.second);
            }
        }
    }

    int steps = 0;
    while (!work_list.empty()) {
        if (steps % 100000 == 0) {
            cerr << "Steps: " << steps << ", Worklist: " << work_list.size() << endl;
        }
        steps += 1;
        auto e = work_list.front();
        work_list.pop();
        tmp_gr->add_edge(e);

        int source = e->get_source();
        int target = e->get_target();
        int symbol = e->get_symbol();

        assert(!is_paren(symbol));

        if (symbol == GRAM_S0) {
            // S -> @S0S0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                addS(work_list, tmp_gr, source, forward->get_target());
            }
            // S -> S0@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                addS(work_list, tmp_gr, backward->get_source(), target);
            }
            // S -> S@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                addS(work_list, tmp_gr, backward->get_source(), target);
            }
            // S -> @S0S
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                addS(work_list, tmp_gr, source, forward->get_target());
            }*/
            // L -> L@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_L)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }
            // L -> @S0L
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }*/
            // L -> @S0L0
            for (auto forward : tmp_gr->edges_from(target, GRAM_L0)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }
            // L -> L0@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_L0)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }
            // R -> R@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }
            // R -> @S0R
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_R)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }*/
            // R -> @S0R0
            for (auto forward : tmp_gr->edges_from(target, GRAM_R0)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }
            // R -> R0@S0
            for (auto backward : tmp_gr->edges_to(source, GRAM_R0)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }
            // S0 -> {i@S0}i
            if (record[source].count(target) != 0) {
                auto &pairs = record[source][target];
                for (auto &p : pairs) {
                    addS0(work_list, tmp_gr, p.first, p.second);
                }
            }
        } else if (symbol == GRAM_L0) {
            // L -> @L0L0
            for (auto forward : tmp_gr->edges_from(target, GRAM_L0)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }
            // L -> L0@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_L0)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }
            // L -> L@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_L)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }
            // L -> @L0L
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_L)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }*/
            // L -> S@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }
            // L -> @L0S
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }*/
            // L -> S0@L0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }
            // L -> @L0S0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }
        } else if (symbol == GRAM_R0) {
            // R -> @R0R0
            for (auto forward : tmp_gr->edges_from(target, GRAM_R0)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }
            // R -> R0@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_R0)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }
            // R -> R@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_R)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }
            // R -> @R0R
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_R)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }*/
            // R -> S@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }
            // R -> @R0S
            /*for (auto forward : tmp_gr->edges_from(target, GRAM_S)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }*/
            // R -> S0@R0
            for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }
            // R -> @R0S0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }
        } else if (symbol == GRAM_S) {
            // S -> @SS0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                addS(work_list, tmp_gr, source, forward->get_target());
            }
            // S -> S0@S
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                addS(work_list, tmp_gr, backward->get_source(), target);
            }*/
            // L -> @SL0
            for (auto forward : tmp_gr->edges_from(target, GRAM_L0)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }
            // L -> L0@S
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_L0)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }*/
            // R -> @SR0
            for (auto forward : tmp_gr->edges_from(target, GRAM_R0)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }
            // R -> R0@S
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_R0)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }*/
            // S0 -> {i@S}i
            if (record[source].count(target) != 0) {
                auto &pairs = record[source][target];
                for (auto &p : pairs) {
                    addS0(work_list, tmp_gr, p.first, p.second);
                }
            }
        } else if (symbol == GRAM_L) {
            // L -> @LL0
            for (auto forward : tmp_gr->edges_from(target, GRAM_L0)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }
            // L -> L0@L
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_L0)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }*/
            // L -> @LS0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                addL(work_list, tmp_gr, source, forward->get_target());
            }
            // L -> S0@L
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                addL(work_list, tmp_gr, backward->get_source(), target);
            }*/
        } else if (symbol == GRAM_R) {
            // R -> @RR0
            for (auto forward : tmp_gr->edges_from(target, GRAM_R0)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }
            // R -> R0@R
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_R0)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }*/
            // R -> @RS0
            for (auto forward : tmp_gr->edges_from(target, GRAM_S0)) {
                addR(work_list, tmp_gr, source, forward->get_target());
            }
            // R -> S0@R
            /*for (auto backward : tmp_gr->edges_to(source, GRAM_S0)) {
                addR(work_list, tmp_gr, backward->get_source(), target);
            }*/
        } else {
            throw "bad implementation";
        }
    }

    for (auto e : tmp_gr->all_edges()) {
        int src = e->get_source();
        int tgt = e->get_target();
        int sym = e->get_symbol();
        if (sym == GRAM_S0) {
            addS(work_list, tmp_gr, src, tgt);
        } else if (sym == GRAM_L0) {
            addL(work_list, tmp_gr, src, tgt);
        } else if (sym == GRAM_R0) {
            addR(work_list, tmp_gr, src, tgt);
        }
    }
    while (!work_list.empty()) {
        auto e = work_list.front();
        work_list.pop();
        tmp_gr->add_edge(e);
    }

    for (int n = 0; n < tmp_gr->get_num_nodes(); n++) {
        for (auto e1 : tmp_gr->edges_to(n, GRAM_R)) {
            for (auto e2 : tmp_gr->edges_from(n, GRAM_L)) {
                int src = e1->get_source();
                int tgt = e2->get_target();
                if (!tmp_gr->is_redundant(src, tgt, GRAM_D)) {
                    auto new_edge = new edge(src, tgt, GRAM_D);
                    tmp_gr->add_record(new_edge);
                    tmp_gr->add_edge(new_edge);
                }
            }
        }
    }

    auto end = clock();
    cerr << "Analyze time: " << (end - start) * 1.0 / CLOCKS_PER_SEC << endl;

    return tmp_gr;
}