//
// Created by Wayne on 3/17/16.
//

#include <iostream>
#include <sys/time.h>
#include <sys/resource.h>
#include "application_graph.h"
#include "analyze_whole.h"

using namespace std;

/*
 * argv[1] : benchmark.txt
 * argv[2] : LibraryMaxNumber.dat
 * argv[3] : evaluation.log
 */
int main(int argc, char *argv[]) {
    ios::sync_with_stdio(false);

    Logger logger(argv[3]);

    logger.log("Evaluation on Whole-Program Analysis\n");
    logger.log("====================================\n");

    ifstream threshold(argv[2]);
    int lib_nodes;
    threshold >> lib_nodes;

    ifstream benchmark(argv[1]);
    long last_mem = 0;

    for (string subject_dat, implemented_dat; benchmark >> subject_dat >> implemented_dat; ) {
        cerr << "Loading " << subject_dat << " ..." << endl;
        const application_graph *app_gr = application_graph::load(subject_dat);
        cerr << "Load done." << endl;

        logger.log("----\n");
        logger.log("Subject: ");
        logger.log(subject_dat);
        logger.log("\n");

        cerr << "Analyzing " << subject_dat << " ..." << endl;
        const application_graph *res_gr = analyze_whole(app_gr, logger);
        cerr << "Analyze done." << endl;

        struct rusage usage;
        getrusage(RUSAGE_SELF, &usage);
        cerr << "Analyze memory: " << (usage.ru_maxrss - last_mem) / 1024.0 << endl;

        logger.log("Memory of Analysis: ");
        logger.log((usage.ru_maxrss - last_mem) / 1024.0);
        logger.log("\n");
        last_mem = usage.ru_maxrss;

        res_gr->get_num_reach(lib_nodes, logger);

        logger.log("----\n");
    }

    logger.close();

    return 0;
}
