//
// Created by Wayne on 15/12/29.
//

#ifndef REACHABILITY_PREMISE_H
#define REACHABILITY_PREMISE_H


#include <vector>
#include <algorithm>
#include <iostream>
#include <unordered_map>

using std::vector;
using std::ostream;
using std::cerr;
using std::endl;
using std::unordered_map;

class premise_0;
class premise_1;
class premise_2;
class premise_3;
class premise_4;
class premise_5;
class premise_f;

class conditional_edge;

class premise_base {
public:
    virtual premise_base* merge(const premise_base *other) const = 0;
    virtual premise_base* merge_0(const premise_0 *other) const = 0;
    virtual premise_base* merge_1(const premise_1 *other) const = 0;
    virtual premise_base* merge_2(const premise_2 *other) const = 0;
    virtual premise_base* merge_3(const premise_3 *other) const = 0;
    virtual premise_base* merge_4(const premise_4 *other) const = 0;
    virtual premise_base* merge_5(const premise_5 *other) const = 0;
    virtual premise_base* merge_f(const premise_f *other) const = 0;

    virtual bool is_covered_by(const premise_base *other) const = 0;
    bool can_cover_0(const premise_0 *other) const { return true; }
    virtual bool can_cover_1(const premise_1 *other) const = 0;
    virtual bool can_cover_2(const premise_2 *other) const = 0;
    virtual bool can_cover_3(const premise_3 *other) const = 0;
    virtual bool can_cover_4(const premise_4 *other) const = 0;
    virtual bool can_cover_5(const premise_5 *other) const = 0;
    virtual bool can_cover_f(const premise_f *other) const = 0;

    virtual size_t size() const = 0;
    virtual size_t hash() const = 0;
    virtual void print() = 0;
    virtual ~premise_base() {}

    virtual void add(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const = 0;
    virtual void add2(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const = 0;
    virtual size_t get_one(int index) const = 0;
    virtual vector<size_t> subsets() const = 0;
};

class premise_0 : public premise_base {
public:
    vector<size_t> subsets() const {
        return vector<size_t>{0};
    }
    size_t get_one(int index) const { throw "bad implementation"; }
    void print() {
        cerr << 0 << endl;
    }
    size_t size() const { return 0; }
    size_t hash() const { return 0; }
    premise_0();
    premise_base* merge(const premise_base *other) const;
    premise_base* merge_0(const premise_0 *other) const;
    premise_base* merge_1(const premise_1 *other) const;
    premise_base* merge_2(const premise_2 *other) const;
    premise_base* merge_3(const premise_3 *other) const;
    premise_base* merge_4(const premise_4 *other) const;
    premise_base* merge_5(const premise_5 *other) const;
    premise_base* merge_f(const premise_f *other) const;

    bool is_covered_by(const premise_base *other) const;
    bool can_cover_1(const premise_1 *other) const { return false; }
    bool can_cover_2(const premise_2 *other) const { return false; }
    bool can_cover_3(const premise_3 *other) const { return false; }
    bool can_cover_4(const premise_4 *other) const { return false; }
    bool can_cover_5(const premise_5 *other) const { return false; }
    bool can_cover_f(const premise_f *other) const { return false; }

    void add(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const {
        ins[1UL << 63].push_back(ce);
    }

    void add2(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const {
        ins[0].push_back(ce);
    }
};

class premise_1 : public premise_base {
    size_t p1;
public:
    vector<size_t> subsets() const {
        return vector<size_t>{0, p1};
    }
    size_t get_one(int index) const { return p1; }
    void print() {
        cerr << "1: " << p1 << endl;
    }
    size_t size() const { return 1; }
    size_t hash() const { return p1; }
    size_t get_p1() const;
    premise_1(size_t p1);
    premise_1(const premise_1 *other);
    premise_base* merge(const premise_base *other) const;
    premise_base* merge_0(const premise_0 *other) const;
    premise_base* merge_1(const premise_1 *other) const;
    premise_base* merge_2(const premise_2 *other) const;
    premise_base* merge_3(const premise_3 *other) const;
    premise_base* merge_4(const premise_4 *other) const;
    premise_base* merge_5(const premise_5 *other) const;
    premise_base* merge_f(const premise_f *other) const;

    bool is_covered_by(const premise_base *other) const;
    bool can_cover_1(const premise_1 *other) const;
    bool can_cover_2(const premise_2 *other) const { return false; }
    bool can_cover_3(const premise_3 *other) const { return false; }
    bool can_cover_4(const premise_4 *other) const { return false; }
    bool can_cover_5(const premise_5 *other) const { return false; }
    bool can_cover_f(const premise_f *other) const { return false; }

    void add(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const {
        ins[p1].push_back(ce);
    }

    void add2(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const {
        ins[p1].push_back(ce);
    }
};

class premise_2 : public premise_base {
    size_t p1, p2;
public:
    vector<size_t> subsets() const {
        return vector<size_t>{0, p1, p2, p1 * 99991UL + p2};
    }
    size_t get_one(int index) const { return (index == 0 ? p1 : p2); }
    void print() {
        cerr << "2: " << p1 << " " << p2 << endl;
    }
    size_t size() const { return 2; }
    size_t hash() const { return p1 * 99991UL + p2; }
    size_t get_p1() const;
    size_t get_p2() const;
    premise_2(size_t p1, size_t p2);
    premise_2(const premise_2 *other);
    premise_base* merge(const premise_base *other) const;
    premise_base* merge_0(const premise_0 *other) const;
    premise_base* merge_1(const premise_1 *other) const;
    premise_base* merge_2(const premise_2 *other) const;
    premise_base* merge_3(const premise_3 *other) const;
    premise_base* merge_4(const premise_4 *other) const;
    premise_base* merge_5(const premise_5 *other) const;
    premise_base* merge_f(const premise_f *other) const;

    bool is_covered_by(const premise_base *other) const;
    bool can_cover_1(const premise_1 *other) const;
    bool can_cover_2(const premise_2 *other) const;
    bool can_cover_3(const premise_3 *other) const { return false; }
    bool can_cover_4(const premise_4 *other) const { return false; }
    bool can_cover_5(const premise_5 *other) const { return false; }
    bool can_cover_f(const premise_f *other) const { return false; }

    void add(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const {
        ins[p1].push_back(ce);
        ins[p2].push_back(ce);
    }

    void add2(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const {
        ins[p1 * 99991UL + p2].push_back(ce);
    }
};

class premise_3 : public premise_base {
    size_t p1, p2, p3;
public:
    vector<size_t> subsets() const {
        return vector<size_t>{0, p1, p2, p3, p1 * 99991UL + p2, p1 * 99991UL + p3, p2 * 99991UL + p3, (p1 * 99991UL + p2) * 99991UL + p3};
    }
    size_t get_one(int index) const { return (index == 0 ? p1 : (index == 1 ? p2 : p3)); }
    void print() {
        cerr << "3: " << p1 << " " << p2 << " " << p3 << endl;
    }
    size_t size() const { return 3; }
    size_t hash() const { return (p1 * 99991UL + p2) * 99991UL + p3; }
    size_t get_p1() const;
    size_t get_p2() const;
    size_t get_p3() const;
    premise_3(size_t p1, size_t p2, size_t p3);
    premise_3(const premise_3 *other);
    premise_base* merge(const premise_base *other) const;
    premise_base* merge_0(const premise_0 *other) const;
    premise_base* merge_1(const premise_1 *other) const;
    premise_base* merge_2(const premise_2 *other) const;
    premise_base* merge_3(const premise_3 *other) const;
    premise_base* merge_4(const premise_4 *other) const;
    premise_base* merge_5(const premise_5 *other) const;
    premise_base* merge_f(const premise_f *other) const;

    bool is_covered_by(const premise_base *other) const;
    bool can_cover_1(const premise_1 *other) const;
    bool can_cover_2(const premise_2 *other) const;
    bool can_cover_3(const premise_3 *other) const;
    bool can_cover_4(const premise_4 *other) const { return false; }
    bool can_cover_5(const premise_5 *other) const { return false; }
    bool can_cover_f(const premise_f *other) const { return false; }

    void add(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const {
        ins[p1].push_back(ce);
        ins[p2].push_back(ce);
        ins[p3].push_back(ce);
    }

    void add2(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const {
        ins[(p1 * 99991UL + p2) * 99991UL + p3].push_back(ce);
    }
};

class premise_4 : public premise_base {
    size_t p1, p2, p3, p4;
public:
    vector<size_t> subsets() const {
        return vector<size_t>{0, p1, p2, p3, p4,
                              p1 * 99991UL + p2, p1 * 99991UL + p3, p1 * 99991UL + p4, p2 * 99991UL + p3,
                              p2 * 99991UL + p4, p3 * 99991UL + p4,
                              (p1 * 99991UL + p2) * 99991UL + p3, (p1 * 99991UL + p2) * 99991UL + p4,
                              (p1 * 99991UL + p3) * 99991UL + p4, (p2 * 99991UL + p3) * 99991UL + p4,
                              ((p1 * 99991UL + p2) * 99991UL + p3) * 99991UL + p4
        };
    }
    size_t get_one(int index) const {
        switch (index) {
            case 0:
                return p1;
            case 1:
                return p2;
            case 2:
                return p3;
            case 3:
                return p4;
            default:
                throw "bad implementation";
        }
    }
    void print() {
        cerr << "4: " << p1 << " " << p2 << " " << p3 << " " << p4 << endl;
    }
    size_t size() const { return 4; }
    size_t hash() const { return ((p1 * 99991UL + p2) * 99991UL + p3) * 99991UL + p4; }
    size_t get_p1() const;
    size_t get_p2() const;
    size_t get_p3() const;
    size_t get_p4() const;
    premise_4(size_t p1, size_t p2, size_t p3, size_t p4);
    premise_4(const premise_4 *other);
    premise_base* merge(const premise_base *other) const;
    premise_base* merge_0(const premise_0 *other) const;
    premise_base* merge_1(const premise_1 *other) const;
    premise_base* merge_2(const premise_2 *other) const;
    premise_base* merge_3(const premise_3 *other) const;
    premise_base* merge_4(const premise_4 *other) const;
    premise_base* merge_5(const premise_5 *other) const;
    premise_base* merge_f(const premise_f *other) const;

    bool is_covered_by(const premise_base *other) const;
    bool can_cover_1(const premise_1 *other) const;
    bool can_cover_2(const premise_2 *other) const;
    bool can_cover_3(const premise_3 *other) const;
    bool can_cover_4(const premise_4 *other) const;
    bool can_cover_5(const premise_5 *other) const { return false; }
    bool can_cover_f(const premise_f *other) const { return false; }

    void add(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const {
        ins[p1].push_back(ce);
        ins[p2].push_back(ce);
        ins[p3].push_back(ce);
        ins[p4].push_back(ce);
    }

    void add2(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const {
        ins[((p1 * 99991UL + p2) * 99991UL + p3) * 99991UL + p4].push_back(ce);
    }
};

class premise_5 : public premise_base {
    size_t p1, p2, p3, p4, p5;
public:
    vector<size_t> subsets() const {
        return vector<size_t>{0, p1, p2, p3, p4, p5,
                              p1 * 99991UL + p2, p1 * 99991UL + p3, p1 * 99991UL + p4, p1 * 99991UL + p5, p2 * 99991UL + p3,
                              p2 * 99991UL + p4, p2 * 99991UL + p5, p3 * 99991UL + p4, p3 * 99991UL + p5, p4 * 99991UL + p5,
                              (p1 * 99991UL + p2) * 99991UL + p3, (p1 * 99991UL + p2) * 99991UL + p4, (p1 * 99991UL + p2) * 99991UL + p5,
                              (p1 * 99991UL + p3) * 99991UL + p4, (p1 * 99991UL + p3) * 99991UL + p5, (p1 * 99991UL + p4) * 99991UL + p5,
                              (p2 * 99991UL + p3) * 99991UL + p4, (p2 * 99991UL + p3) * 99991UL + p5, (p2 * 99991UL + p4) * 99991UL + p5,
                              (p3 * 99991UL + p4) * 99991UL + p5,
                              ((p1 * 99991UL + p2) * 99991UL + p3) * 99991UL + p4, ((p1 * 99991UL + p2) * 99991UL + p3) * 99991UL + p5,
                              ((p1 * 99991UL + p2) * 99991UL + p4) * 99991UL + p5, ((p1 * 99991UL + p3) * 99991UL + p4) * 99991UL + p5,
                              ((p2 * 99991UL + p3) * 99991UL + p4) * 99991UL + p5,
                              (((p1 * 99991UL + p2) * 99991UL + p3) * 99991UL + p4) * 99991UL + p5
        };
    }
    size_t get_one(int index) const {
        switch (index) {
            case 0:
                return p1;
            case 1:
                return p2;
            case 2:
                return p3;
            case 3:
                return p4;
            case 4:
                return p5;
            default:
                throw "bad implementation";
        }
    }
    void print() {
        cerr << "5: " << p1 << " " << p2 << " " << p3 << " " << p4 << " " << p5 << endl;
    }
    size_t size() const { return 5; }
    size_t hash() const { return (((p1 * 99991UL + p2) * 99991UL + p3) * 99991UL + p4) * 99991UL + p5; }
    size_t get_p1() const;
    size_t get_p2() const;
    size_t get_p3() const;
    size_t get_p4() const;
    size_t get_p5() const;
    premise_5(size_t p1, size_t p2, size_t p3, size_t p4, size_t p5);
    premise_5(const premise_5 *other);
    premise_base* merge(const premise_base *other) const;
    premise_base* merge_0(const premise_0 *other) const;
    premise_base* merge_1(const premise_1 *other) const;
    premise_base* merge_2(const premise_2 *other) const;
    premise_base* merge_3(const premise_3 *other) const;
    premise_base* merge_4(const premise_4 *other) const;
    premise_base* merge_5(const premise_5 *other) const;
    premise_base* merge_f(const premise_f *other) const;

    bool is_covered_by(const premise_base *other) const;
    bool can_cover_1(const premise_1 *other) const;
    bool can_cover_2(const premise_2 *other) const;
    bool can_cover_3(const premise_3 *other) const;
    bool can_cover_4(const premise_4 *other) const;
    bool can_cover_5(const premise_5 *other) const;
    bool can_cover_f(const premise_f *other) const { return false; }

    void add(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const {
        ins[p1].push_back(ce);
        ins[p2].push_back(ce);
        ins[p3].push_back(ce);
        ins[p4].push_back(ce);
        ins[p5].push_back(ce);
    }

    void add2(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const {
        ins[(((p1 * 99991UL + p2) * 99991UL + p3) * 99991UL + p4) * 99991UL + p5].push_back(ce);
    }
};

class premise_f : public premise_base {
    vector<size_t> premise;
    size_t record_hash;
public:
    vector<size_t> subsets() const {
        int sz = premise.size();
        int upper = 1 << sz;
        vector<size_t> res(upper);
        for (int msk = 1; msk < upper; msk++) {
            int tmp = msk & -msk;
            int p = __builtin_ctz(tmp);
            res[msk] = res[msk ^ tmp] * 99991UL + premise.at(sz - 1 - p);
        }
        return res;
    }
    size_t get_one(int index) const { return premise.at(index); }
    void print() {
        cerr << premise.size() << endl;
    }
    size_t size() const { return premise.size(); }
    size_t hash() const { return record_hash; }
    const vector<size_t>& get_premise() const;
    premise_f(const vector<size_t> &premise);
    premise_f(const premise_f *other);
    premise_base* merge(const premise_base *other) const;
    premise_base* merge_0(const premise_0 *other) const;
    premise_base* merge_1(const premise_1 *other) const;
    premise_base* merge_2(const premise_2 *other) const;
    premise_base* merge_3(const premise_3 *other) const;
    premise_base* merge_4(const premise_4 *other) const;
    premise_base* merge_5(const premise_5 *other) const;
    premise_base* merge_f(const premise_f *other) const;

    bool is_covered_by(const premise_base *other) const;
    bool can_cover_1(const premise_1 *other) const;
    bool can_cover_2(const premise_2 *other) const;
    bool can_cover_3(const premise_3 *other) const;
    bool can_cover_4(const premise_4 *other) const;
    bool can_cover_5(const premise_5 *other) const;
    bool can_cover_f(const premise_f *other) const;

    void add(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const {
        for (size_t p : premise) {
            ins[p].push_back(ce);
        }
    }

    void add2(unordered_map<size_t, vector<const conditional_edge*>> &ins, const conditional_edge *ce) const {
        ins[record_hash].push_back(ce);
    }
};


#endif //REACHABILITY_PREMISE_H
